; 
    include \masm32\include\masm32rt.inc
; 

comment * -----------------------------------------------------
                          Build this code with
                       "CONSOLE ASSEMBLE AND LINK"

        the procedure call tree for this app is as follows.

        main ---,
                |
            line_read ------ lfcnt
                |
            line_parse ----- test_word
                |
           expand_macro -,-- parse_code
                         |
                         '-- remove$
                         |
                         '-- wreplace$

        ----------------------------------------------------- *

    line_read       PROTO :DWORD
    lfcnt           PROTO :DWORD
    line_parse      PROTO :DWORD,:DWORD
    test_word       PROTO :DWORD
    get_macro       PROTO :DWORD
    expand_macro    PROTO :DWORD,:DWORD,:DWORD,:DWORD
    parse_code      PROTO :DWORD,:DWORD

    wreplace$ MACRO src,dst,wrd1,wrd2
      invoke wordreplace,src,dst,wrd1,wrd2
      invoke szCopy,dst,src
      mov eax, src
      EXITM <eax>
    ENDM

    .data?
      array0    dd ?
      array1    dd ?                        ; macro expansion arrays
      amem0     dd ?
      amem1     dd ?

      buf128k   dd ?                        ; whole macro buffer handle
      mac128k   dd ?

    .code

    include testword.asm

start:
  
; 

  ; ----------------------------------
  ; allocate array for macro arguments
  ; ----------------------------------
    invoke create_array,128,128  
    mov array0, eax                         ; string array address
    mov amem0, ecx                          ; block memory address
  ; ----------------------
  ; macro expansion arrays
  ; ----------------------
    invoke create_array,128,128             ; src line array
    mov array1, eax                         ; address of array
    mov amem1, ecx                          ; array main memory address

  ; ----------------------
    mov buf128k, alloc(1024*128)
    mov mac128k, alloc(1024*128)
  ; ----------------------

    call main                               ; call the main procedure

  ; ----------------------
    free buf128k
    free mac128k
  ; ----------------------
    free array0                             ; deallocate memory
    free amem0

    free array1
    free amem1

    exit                                    ; terminate the program

; 

main proc

    LOCAL hMem  :DWORD
    LOCAL hFile :DWORD
    LOCAL flen  :DWORD
    LOCAL mlen  :DWORD
    LOCAL br    :DWORD
    LOCAL pbuf  :DWORD
    LOCAL buffer[260]:BYTE

    mov pbuf, ptr$(buffer)                  ; cast buffer to pointer

    invoke GetCL,1,pbuf                     ; get 1st command line argument

    .if rv(exist,pbuf) == 0                 ; test if input file exists
      print "Cannot find file",13,10        ; print error and exit if it does not
      ret
    .endif

    mov hMem, InputFile(pbuf)               ; load file into buffer

    invoke line_read,hMem                   ; cal the line reader proc

    free hMem

    ret

main endp

; 

align 4

lfcnt proc txt:DWORD

  ; ---------------------------------------------
  ; get line count of text by counting line feeds
  ; ---------------------------------------------
    xor eax, eax                            ; set counter to zero
    mov ecx, txt
    sub ecx, 1
    jmp @F

  pre:
    add eax, 1                              ; add 1 to return value if LF found
  @@:
    add ecx, 1
    cmp BYTE PTR [ecx], 0                   ; test for zero terminator
    je ccout
    cmp BYTE PTR [ecx], 10                  ; test line feed
    je pre
    jmp @B

  ccout:

    ret

lfcnt endp

; 

align 4
; -------------------------------------
; read the source file 1 line at a time
; -------------------------------------
line_read proc src:DWORD

    LOCAL spos  :DWORD                      ; starting position variable
    LOCAL pbuf  :DWORD                      ; pointer for buffer
    LOCAL lcnt  :DWORD                      ; line count
    LOCAL lchar :DWORD                      ; left character of tested line
    LOCAL bw    :DWORD                      ; bytes written return value
    LOCAL lnum  :DWORD                      ; line number in source file

    LOCAL buffer[1024]:BYTE

    mov pbuf, ptr$(buffer)                  ; cast buffer to a pointer
    mov spos, 0                             ; set location pointer to zero
    mov lnum, 0                             ; set line number counter to zero
    mov lcnt, rv(lfcnt,src)                 ; get the line count
    add lcnt, 2                             ; preload by 1 to get last line

  @@:
    add lnum, 1
    sub lcnt, 1
    jz bypass                               ; exit loop if line count zero
    mov spos, linein$(src,pbuf,spos)        ; read line into memory at pbuf
    mov lchar, rv(tstline,pbuf)             ; test line
    .if lchar != 0 && lchar != ";"          ; if not empty or a comment
      invoke line_parse,pbuf,lnum           ; tokenise & display result
      test eax, eax                         ; exit if error is returned
      jz bypass
    .else
      print pbuf,13,10                      ; else diaplay what was there
    .endif
    jmp @B

  bypass:

    ret

line_read endp

; 

align 4
; ------------------------------------
; parse and test each line for content
; ------------------------------------
line_parse proc src:DWORD,lnum:DWORD

    LOCAL wcnt  :DWORD                      ; word count return value
    LOCAL bw    :DWORD                      ; bytes written return value
    LOCAL tval  :DWORD                      ; address of macro text
    LOCAL mlen  :DWORD                      ; length variable for macro text
    LOCAL var   :DWORD                      ; temp memory address

    push esi

  ; -------------------------------
  ; parse line, load tokens into
  ; array and return count of words
  ; -------------------------------
    mov wcnt, rv(parse_line,src,array0)
    mov esi, array0                         ; array address in ESI

    mov tval, rv(test_word,[esi])           ; test if word is macro

    .if tval == 0
      print src,13,10                       ; if word not listed print source
    .else
      mov var, rv(expand_macro,src,tval,buf128k,lnum)
      test eax, eax                         ; exit on error
      jz quit
      print var,13,10                       ; else print expanded macro
      mov eax, 1
    .endif

  quit:

    pop esi

    ret

line_parse endp

; 

align 4
; -----------------------------------------
; expand macro and return address of result
; -----------------------------------------
expand_macro proc src:DWORD,mac:DWORD,dst:DWORD,lnum:DWORD

    LOCAL cnt1      :DWORD                  ; source arg count
    LOCAL cnt2      :DWORD                  ; macro arg count
    LOCAL spos      :DWORD                  ; start pos for line
    LOCAL lpmac     :DWORD                  ; local pointer for global address

    LOCAL pbuf  :DWORD                      ; line buffer pointer
    LOCAL buffer[1024]:BYTE                 ; line buffer

    push ebx
    push esi
    push edi

    mov pbuf, ptr$(buffer)                  ; cast buffer to pointer
    mov spos, 0                             ; start pos is zero
    mov spos, linein$(mac,pbuf,spos)        ; get 1st line from macro
    invoke szCopy,mac,mac128k               ; copy macro to buffer
    mov cnt1, rv(parse_code,src,array0)     ; parse the src line
    mov cnt2, rv(parse_code,pbuf,array1)    ; parse the 1st macro line

    sub eax, 1                              ; correct the arg count (macro line has 1 extra)
    cmp cnt1, eax                           ; compare argument counts
    je @F
  ; ---------------------------------------------
  ; Display error and exit if counts do not match
  ; ---------------------------------------------
    print "Parameter count error =",62,62,32
    print src,13,10,"*** Line Number "
    print str$(lnum)," ***",13,10
    xor eax, eax
    jmp quit
  ; ---------------------------------------------
  @@:

    mov ebx, cnt2                           ; use the MACRO arg count as reference
    sub ebx, 2                              ; correct arg count in EBX

    mov esi, array0
    add esi, 4                              ; skip 1st word on src line

    mov edi, array1
    add edi, 8                              ; skip 1st and 2nd words in macro line

    mov eax, mac128k                        ; copy global to LOCAL pointer
    mov lpmac, eax

    mov eax, spos
    add lpmac, eax                          ; mov start of macro past 1st line

  ; ----------------------
  ; remove trailing "ENDM"
  ; ----------------------
    mov lpmac, remove$(lpmac,chr$(13,10,"ENDM"))

  @@:
    mov lpmac, wreplace$(lpmac,dst,[edi],[esi])
    add esi, 4
    add edi, 4
    sub ebx, 1
    jnz @B

    mov eax, lpmac                          ; return buffer address of result

  quit:
    pop edi
    pop esi
    pop ebx

    ret

expand_macro endp

; 

.data
  align 4
  ctbl \
    db 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,1,0,0,0,0,0,0,0,1,1,0,1,1,0
    db 1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0
    db 0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
    db 1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1
    db 0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
    db 1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

    ;   allowable character range
    ;   -------------------------
    ;   upper & lower case characters
    ;   numbers
    ;   " "           ; quotation marks
    ;   [ ]           ; address enclosure
    ;   + - *         ; displacement and multiplier calculation
    ;   : .           ; punctuation

.code

; 

align 4

parse_code proc src:DWORD,array:DWORD

    push esi
    push edi

    mov esi, src                        ; line in ESI
    mov edi, array                      ; word array in EDI
    mov ecx, [edi]                      ; 1st array member address in ECX
    xor eax, eax                        ; clear EAX
    xor edx, edx                        ; set arg counter to zero

  ; ----------------------------------
  align 4
  badchar:
    mov al, [esi]
    add esi, 1
    test al, al                         ; test if byte is terminator
    jz gaout
    cmp BYTE PTR [eax+ctbl], 1          ; is it a good char ?
    jne badchar
    add edx, 1                          ; add to arg counter
    jmp writechar
  goodchar:
    mov al, [esi]
    add esi, 1
    test al, al                         ; test if byte is terminator
    jz gaout
    cmp BYTE PTR [eax+ctbl], 0          ; is it a bad char ?
    je reindex
  writechar:
    cmp al, "["                         ; test for opening square bracket
    je wsqb                             ; branch to handler if it is
    cmp al, 34                          ; test for opening quotation
    je preq                             ; branch to quote handler if it is
    mov [ecx], al                       ; write byte to buffer
    add ecx, 1
    jmp goodchar
  ; ----------------------------------
  align 4
  reindex:
    mov BYTE PTR [ecx], 0               ; write terminator to array member
    add edi, 4                          ; index to next array member
    mov ecx, [edi]                      ; put its address in ECX
    jmp badchar
  ; ----------------------------------
  align 4
  preq:
    mov [ecx], al                       ; write byte to buffer
    add ecx, 1
  quotes:                               ; quotation handler
    mov al, [esi]
    add esi, 1
    test al, al                         ; test if byte is terminator
    jz qterror                          ; jump to quotation error
 ;     cmp al, 32
 ;     je quotes
  wqot:
    mov [ecx], al                       ; write byte to buffer
    add ecx, 1
    cmp al, 34                          ; test for closing quote
    je reindex
    jmp quotes

  align 4
  squareb:                              ; square bracket handler
    mov al, [esi]
    add esi, 1
    test al, al                         ; test if byte is terminator
    jz sberror                          ; jump to square bracket error
    cmp al, 32
    je squareb
  wsqb:
    mov [ecx], al                       ; write byte to buffer
    add ecx, 1
    cmp al, "]"
    je reindex
    jmp squareb
  ; ----------------------------------
  align 4
  qterror:                              ; quotation error
    mov ecx, 2                          ; set ECX as 2
    jmp gaquit                          ; no closing quoation

  align 4
  sberror:                              ; square bracket error
    mov ecx, 1                          ; set ECX to 1
    jmp gaquit                          ; no closing square bracket

  align 4
  gaout:
    mov BYTE PTR [ecx], 0               ; terminate last buffer written to.
    add edi, 4
    mov ecx, [edi]
    mov BYTE PTR [ecx], 0               ; set next buffer with leading zero
    xor ecx, ecx                        ; ECX set to zero is NO ERROR
  gaquit:
    mov eax, edx                        ; return arg count

    pop edi
    pop esi

    ret

parse_code endp

; 

end start                                   ; end of code

